package database

import (
	"database/sql"
	"errors"
	"strconv"

	_ "github.com/go-sql-driver/mysql"
	"github.com/pivotal-cf-experimental/mysql-metrics/metrics"
	"strings"
)

type DatabaseClient struct {
	connection *sql.DB
}

func NewDatabaseClient(connection *sql.DB) *DatabaseClient {
	return &DatabaseClient{
		connection: connection,
	}
}

func (dc *DatabaseClient) QueryMetrics(metricMappings map[string]*metrics.Metric) ([]*metrics.Metric, error) {
	rows, err := dc.connection.Query("SHOW GLOBAL STATUS")
	if err != nil {
		return []*metrics.Metric{dc.availableMetric(false)}, err
	}
	var gatheredMetrics []*metrics.Metric
	gatheredMetrics = append(gatheredMetrics, dc.addMetrics(rows, metricMappings)...)
	rows.Close()

	rows, err = dc.connection.Query("SHOW GLOBAL VARIABLES")
	if err != nil {
		return []*metrics.Metric{dc.availableMetric(false)}, err
	}
	gatheredMetrics = append(gatheredMetrics, dc.addMetrics(rows, metricMappings)...)
	rows.Close()

	return append(gatheredMetrics, dc.availableMetric(true)), nil
}

func (dc *DatabaseClient) addMetrics(rows *sql.Rows, metricMappings map[string]*metrics.Metric) []*metrics.Metric {
	var gatheredMetrics []*metrics.Metric

	for rows.Next() {
		var key, value string

		rows.Scan(&key, &value)
		spec := metricMappings[strings.ToLower(key)]
		if spec != nil {
			metric := metrics.Metric{
				Key:      spec.Key,
				Unit:     spec.Unit,
				RawValue: value,
			}
			floatValue, err := dc.parseFloat(value)
			if err != nil {
				floatValue, err = dc.parseOnOff(value)
			}
			if err != nil {
				floatValue, err = dc.parseClusterStatus(value)
			}
			if err != nil {
				metric.HasError = true
			}
			metric.Value = floatValue
			gatheredMetrics = append(gatheredMetrics, &metric)
		}
	}
	return gatheredMetrics
}

func (dc *DatabaseClient) availableMetric(isAvailable bool) *metrics.Metric {
	availableAsFloat := 0.0
	if isAvailable {
		availableAsFloat = 1.0
	}
	return &metrics.Metric{Key: "/p-mysql/available", Value: availableAsFloat, Unit: "bool"}
}

func (dc *DatabaseClient) parseFloat(value string) (float64, error) {
	return strconv.ParseFloat(value, 64)
}

func (dc *DatabaseClient) parseOnOff(value string) (float64, error) {
	var floatValue float64
	var err error

	switch strings.ToLower(value) {
	case "on":
		floatValue = 1.0
	case "off":
		floatValue = 0.0
	default:
		err = errors.New("could not convert wsrep_ready")
	}

	return floatValue, err
}

func (dc *DatabaseClient) parseClusterStatus(value string) (float64, error) {
	var floatValue float64
	var err error

	switch strings.ToLower(value) {
	case "primary":
		floatValue = 1
	case "non-primary":
		floatValue = 0
	case "disconnected":
		floatValue = -1
	default:
		err = errors.New("could not convert wsrep_cluster_status")
	}

	return floatValue, err
}
