package database_test

import (
	"database/sql"
	"errors"
	"fmt"

	"github.com/DATA-DOG/go-sqlmock"
	"github.com/pivotal-cf-experimental/mysql-metrics/database"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf-experimental/mysql-metrics/metrics"
	"strings"
)

var _ = Describe("DatabaseClient", func() {

	var (
		conn           *sql.DB
		mock           sqlmock.Sqlmock
		err            error
		client         *database.DatabaseClient
		metricMappings map[string]*metrics.Metric
	)

	findMetricByKey := func(metrics []*metrics.Metric, key string) *metrics.Metric {
		for i := range metrics {
			aMetric := metrics[i]
			if strings.Contains(aMetric.Key, key) {
				return aMetric
			}
		}
		return nil
	}

	seedEmptyGlobalVariablesResponse := func() {
		mock.ExpectQuery(`SHOW GLOBAL VARIABLES`).WillReturnRows(sqlmock.NewRows([]string{"variable_name", "value"}))
	}

	seedOneRowFromDatabase := func(keyName string, value string) {
		row := sqlmock.NewRows([]string{"variable_name", "value"}).AddRow(keyName, value)
		mock.ExpectQuery(`SHOW GLOBAL STATUS`).WillReturnRows(row)
		seedEmptyGlobalVariablesResponse()
	}

	BeforeEach(func() {
		metricMappings = map[string]*metrics.Metric{
			"questions":                     {Key: "/p-mysql/global_statuses/questions", Unit: "metric"},
			"innodb_buffer_pool_pages_free": {Key: "/p-mysql/innodb/buffer_pool_pages_free", Unit: "byte"},
			"innodb_data_read":              {Key: "/p-mysql/innodb/data_read", Unit: "byte"},
			"wsrep_cluster_status":          {Key: "/p-mysql/galera/wsrep_cluster_status", Unit: "number"},
			"wsrep_ready":                   {Key: "/p-mysql/galera/wsrep_ready", Unit: "bool"},
			"max_connections":               {Key: "/p-mysql/variables/max_connections", Unit: "integer"},
		}

		conn, mock, err = sqlmock.New()
		Expect(err).NotTo(HaveOccurred())

		client = database.NewDatabaseClient(conn)
	})

	Describe("when the query errors", func() {
		It("returns an error and database is down metric", func() {
			mock.ExpectQuery(`SHOW GLOBAL STATUS`).WillReturnError(errors.New("query error"))
			resultMetrics, err := client.QueryMetrics(metricMappings)
			Expect(err).To(HaveOccurred())
			Expect(err).To(MatchError("query error"))
			Expect(resultMetrics).NotTo(BeNil())
			Expect(len(resultMetrics)).To(Equal(1))

			metric := resultMetrics[0]
			Expect(metric.Key).To(Equal("/p-mysql/available"))
			Expect(metric.Value).To(Equal(0.0))
			Expect(metric.HasError).To(BeFalse())
		})
	})

	Describe("when the query returns non-float data", func() {
		It("returns the error (and the good metrics)", func() {
			row := sqlmock.NewRows([]string{"variable_name", "value"}).
				AddRow("Questions", "Nope").
				AddRow("Innodb_buffer_pool_pages_free", "0")
			mock.ExpectQuery(`SHOW GLOBAL STATUS`).WillReturnRows(row)
			seedEmptyGlobalVariablesResponse()
			resultMetrics, err := client.QueryMetrics(metricMappings)
			Expect(err).NotTo(HaveOccurred())
			Expect(resultMetrics).NotTo(BeNil())
			Expect(len(resultMetrics)).To(Equal(3))

			metric := resultMetrics[0]
			Expect(metric.Key).To(Equal("/p-mysql/global_statuses/questions"))
			Expect(metric.RawValue).To(Equal("Nope"))
			Expect(metric.HasError).To(BeTrue())

			metric = resultMetrics[1]
			Expect(metric.Key).To(Equal("/p-mysql/innodb/buffer_pool_pages_free"))
			Expect(metric.HasError).To(BeFalse())
		})
	})

	Describe("when the query returns expected data", func() {
		var (
			numQuestions      float64
			dataRead          float64
			maxConnections    float64
			clusterStatus     string
			galeraReadyStatus string
		)
		BeforeEach(func() {
			numQuestions = float64(125146.0)
			dataRead = float64(24121.0)
			clusterStatus = "Primary" // this might also be PRIMARY sometimes?
			galeraReadyStatus = "ON"
			maxConnections = 12
		})

		seedDatabase := func() {
			row := sqlmock.NewRows([]string{"variable_name", "value"}).
				AddRow("something", "doesntmatter").
				AddRow("Questions", fmt.Sprintf("%f", numQuestions)).
				AddRow("Innodb_data_read", fmt.Sprintf("%f", dataRead)).
				AddRow("wsrep_cluster_status", clusterStatus).
				AddRow("wsrep_ready", galeraReadyStatus).
				AddRow("Other Thing", "Still doesn't matter")
			mock.ExpectQuery(`SHOW GLOBAL STATUS`).WillReturnRows(row)

			row = sqlmock.NewRows([]string{"variable_name", "value"}).AddRow("max_connections", fmt.Sprintf("%f", maxConnections))
			mock.ExpectQuery(`SHOW GLOBAL VARIABLES`).WillReturnRows(row)
		}

		It("returns an array of metric objects", func() {
			seedDatabase()
			resultMetrics, err := client.QueryMetrics(metricMappings)

			Expect(err).NotTo(HaveOccurred())
			Expect(resultMetrics).NotTo(BeNil())
			Expect(len(resultMetrics)).To(Equal(6))

			aResultMetric := resultMetrics[0]
			Expect(aResultMetric.Value).To(Equal(numQuestions))
			Expect(aResultMetric.Key).To(Equal("/p-mysql/global_statuses/questions"))
			Expect(aResultMetric.Unit).To(Equal("metric"))

			aResultMetric = resultMetrics[1]
			Expect(aResultMetric.Value).To(Equal(dataRead))
			Expect(aResultMetric.Key).To(Equal("/p-mysql/innodb/data_read"))
			Expect(aResultMetric.Unit).To(Equal("byte"))

			aResultMetric = resultMetrics[len(resultMetrics)-2]
			Expect(aResultMetric.Value).To(Equal(maxConnections))
			Expect(aResultMetric.Key).To(Equal("/p-mysql/variables/max_connections"))
			Expect(aResultMetric.Unit).To(Equal("integer"))

			aResultMetric = resultMetrics[len(resultMetrics)-1]
			Expect(aResultMetric.Value).To(Equal(1.0))
			Expect(aResultMetric.Key).To(Equal("/p-mysql/available"))
			Expect(aResultMetric.Unit).To(Equal("bool"))
		})

		Describe("when recovering from a prior bad parse", func() {
			It("should treat the metric as brand new with no history of the past parse", func() {
				seedOneRowFromDatabase("wsrep_cluster_status", "This Breaks Parse")

				resultMetrics, err := client.QueryMetrics(metricMappings)
				Expect(err).NotTo(HaveOccurred())
				Expect(resultMetrics).NotTo(BeNil())

				metric := resultMetrics[0]
				Expect(metric.RawValue).To(Equal("This Breaks Parse"))
				Expect(metric.HasError).To(BeTrue())

				seedOneRowFromDatabase("wsrep_cluster_status", "Primary")

				resultMetrics, err = client.QueryMetrics(metricMappings)
				Expect(err).NotTo(HaveOccurred())
				Expect(resultMetrics).NotTo(BeNil())

				metric = resultMetrics[0]
				Expect(metric.HasError).To(BeFalse())
				Expect(metric.Value).To(Equal(1.0))
			})
		})

		Describe("when the data needs a special mapping", func() {
			BeforeEach(func() {
				seedDatabase()
			})

			It("should map cluster status to a number", func() {
				resultMetrics, err := client.QueryMetrics(metricMappings)

				Expect(err).NotTo(HaveOccurred())
				Expect(resultMetrics).NotTo(BeNil())

				metric := findMetricByKey(resultMetrics, "wsrep_cluster_status")
				Expect(metric).NotTo(BeNil())
				// Primary maps to 1
				Expect(metric.Value).To(Equal(1.0))
				Expect(metric.Unit).To(Equal("number"))

				seedOneRowFromDatabase("wsrep_cluster_status", "Non-Primary")

				resultMetrics, err = client.QueryMetrics(metricMappings)
				metric = findMetricByKey(resultMetrics, "wsrep_cluster_status")
				Expect(metric).NotTo(BeNil())
				// Non-Primary maps to 0
				Expect(metric.Value).To(Equal(0.0))

				seedOneRowFromDatabase("wsrep_cluster_status", "Disconnected")

				resultMetrics, err = client.QueryMetrics(metricMappings)
				metric = findMetricByKey(resultMetrics, "wsrep_cluster_status")
				Expect(metric).NotTo(BeNil())
				// Disconnected maps to -1
				Expect(metric.Value).To(Equal(-1.0))

				seedOneRowFromDatabase("wsrep_cluster_statuS", "DISCONNECTED")

				resultMetrics, err = client.QueryMetrics(metricMappings)
				metric = findMetricByKey(resultMetrics, "wsrep_cluster_status")
				Expect(metric).NotTo(BeNil())
				// Disconnected maps to -1
				Expect(metric.Value).To(Equal(-1.0))
				Expect(metric.HasError).To(BeFalse())
			})

			It("should map wsrep_ready to a 1 or 0", func() {
				resultMetrics, err := client.QueryMetrics(metricMappings)
				Expect(err).NotTo(HaveOccurred())

				metric := findMetricByKey(resultMetrics, "wsrep_ready")
				Expect(metric).NotTo(BeNil())
				// ON to 1
				Expect(metric.Value).To(Equal(1.0))
				Expect(metric.Unit).To(Equal("bool"))

				seedOneRowFromDatabase("wsrep_ready", "OFF")

				resultMetrics, err = client.QueryMetrics(metricMappings)
				metric = findMetricByKey(resultMetrics, "wsrep_ready")
				// OFF to 0
				Expect(metric.Value).To(Equal(0.0))

				seedOneRowFromDatabase("wsrep_ready", "FoobarValue")

				resultMetrics, err = client.QueryMetrics(metricMappings)
				metric = findMetricByKey(resultMetrics, "wsrep_ready")
				Expect(resultMetrics).NotTo(BeNil())
				metric = findMetricByKey(resultMetrics, "wsrep_ready")
				Expect(metric).NotTo(BeNil())
				Expect(metric.RawValue).To(Equal("FoobarValue"))
				Expect(metric.HasError).To(BeTrue())
			})
		})
	})
})
