package main

import (
	"database/sql"
	"encoding/json"
	"flag"
	"fmt"
	"io/ioutil"
	"os"
	"time"

	"log"

	"github.com/cloudfoundry/dropsonde"
	configNamespace "github.com/pivotal-cf-experimental/mysql-metrics/config"
	"github.com/pivotal-cf-experimental/mysql-metrics/database"
	"github.com/pivotal-cf-experimental/mysql-metrics/metrics"
)

const (
	defaultConfigPath              = "/var/vcap/jobs/mysql-metrics/config/mysql-config.yml"
	defaultMetricMappingConfigPath = "/var/vcap/jobs/mysql-metrics/config/metric-mapping-config.yml"
)

var configFilepath = flag.String("c", defaultConfigPath, "location of config file")
var logFilepath = flag.String("l", "", "location of log file")
var mappingFilepath = flag.String("m", defaultMetricMappingConfigPath, "location of metric mapping config file")

func setupLogging() {
	if *logFilepath != "" {
		file, err := os.OpenFile(*logFilepath, os.O_WRONLY|os.O_APPEND|os.O_CREATE, 0660)
		if err != nil {
			log.Printf("Could not open log file: %s\n", err)
			log.SetFlags(0)
			log.SetOutput(ioutil.Discard)
			return
		}
		log.SetOutput(file)
	} else {
		log.SetFlags(0)
		log.SetOutput(ioutil.Discard)
	}
}

func main() {
	flag.Parse()

	setupLogging()

	config := &configNamespace.Config{}
	err := configNamespace.LoadFromFile(*configFilepath, config)

	if err != nil {
		log.Printf("config file is not formatted correctly\n")
		panic(err)
	}

	metricMappingConfig := &configNamespace.MetricMappingConfig{}
	err = configNamespace.LoadFromFile(*mappingFilepath, metricMappingConfig)

	if err != nil {
		log.Printf("metric mapping config file is not formatted correctly\n")
		panic(err)
	}

	err = dropsonde.Initialize("localhost:3457", "p-mysql")
	if err != nil {
		log.Printf("dropsonde failed to initialize\n")
		panic(err)
	}

	conn := Connection(config)
	client := database.NewDatabaseClient(conn)

	metricsWriter := metrics.NewMetricWriter(new(metrics.DropsondeSender))
	process(client, metricsWriter, metricMappingConfig.Mappings)
	metricsInterval := time.Duration(config.MetricsFrequency) * time.Second
	for {
		select {
		case <-time.After(metricsInterval):
			process(client, metricsWriter, metricMappingConfig.Mappings)
		}

	}
}

func process(client *database.DatabaseClient, metricsWriter *metrics.MetricWriter, metricMappings map[string]*metrics.Metric) {
	metrics, err := client.QueryMetrics(metricMappings)

	if err != nil {
		log.Printf("Could not get metrics: %s\n", err)
	}
	log.Printf("Logging %d metrics", len(metrics))
	for i := range metrics {
		metric := metrics[i]
		if metric.HasError {
			log.Printf("Metric had error converting to a float, %v\n", metric)
		} else {
			if err := metricsWriter.Write(metric); err != nil {
				log.Printf("Failed to send metric: %s\n", err)
			}
			value, err := json.Marshal(metric)
			if err != nil {
				log.Printf("Failed to unmarshal metric: %v\n", metric)
			} else {
				log.Printf("%s\n", value)
			}
		}
	}

}

func Connection(config *configNamespace.Config) *sql.DB {
	conn, err := sql.Open("mysql", fmt.Sprintf("%s:%s@tcp(%s:%d)/", config.Username, config.Password, "localhost", 3306))
	if err != nil {
		panic(fmt.Sprintf("Database configuration problem: %v", err))
	}

	return conn
}
