package main_test

import (
	"bytes"
	"fmt"
	"io/ioutil"
	"os"
	"os/exec"
	"path/filepath"
	"time"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/onsi/gomega/gexec"
)

var _ = Describe("Main", func() {
	var (
		configFilepath  string
		mappingFilepath string
		tempDir         string
		password        string
		username        string
		err             error
		session         *gexec.Session
		metricFrequency int
	)

	BeforeEach(func() {
		if env, ok := os.LookupEnv("MYSQL_USER"); ok {
			username = env
		}
		if env, ok := os.LookupEnv("MYSQL_PASSWORD"); ok {
			password = env
		}

		metricFrequency = 1
		tempDir, err = ioutil.TempDir("", "")
		Expect(err).NotTo(HaveOccurred())
		configString := fmt.Sprintf(`{"username":"%s","password":"%s","metrics_frequency":%d,}`, username, password, metricFrequency)
		Expect(err).NotTo(HaveOccurred())

		configFilepath = filepath.Join(tempDir, "metric-config.yml")

		mappingFilepath = "./fixtures/mapping-fixtures.yml"

		err = ioutil.WriteFile(configFilepath, []byte(configString), os.ModePerm)
		Expect(err).NotTo(HaveOccurred())

	})
	AfterEach(func() {
		err = os.RemoveAll(tempDir)
		Expect(err).NotTo(HaveOccurred())
		Eventually(session.Interrupt()).Should(gexec.Exit())
	})

	runMainWithArgs := func(args ...string) {
		args = append(
			args,
			"-c", configFilepath,
			"-m", mappingFilepath,
		)

		_, err := fmt.Fprintf(GinkgoWriter, "Running command: %v\n", args)
		Expect(err).NotTo(HaveOccurred())

		command := exec.Command(metricsBinPath, args...)
		session, err = gexec.Start(command, GinkgoWriter, GinkgoWriter)
		Expect(err).NotTo(HaveOccurred())
	}

	It("is able to run the binary without exit", func() {
		runMainWithArgs()

		Consistently(session.ExitCode, 2*time.Second).Should(Equal(-1))
	})

	Describe("when logging is enabled", func() {

		It("emits multiple log entries", func() {
			logFilePath := filepath.Join(tempDir, "metrics.log")
			runMainWithArgs("-l", logFilePath)
			Consistently(session.ExitCode, time.Duration(metricFrequency)*time.Second).Should(Equal(-1))
			contents, err := ioutil.ReadFile(logFilePath)
			Expect(err).NotTo(HaveOccurred())
			Expect(contents).NotTo(BeEmpty())
			contentsAsString := fmt.Sprintf("%s", contents)
			Expect(contentsAsString).To(ContainSubstring("/p-mysql/innodb/buffer_pool_pages_free"))
			firstLineCount := bytes.Count(contents, []byte("\n"))
			Expect(firstLineCount).Should(BeNumerically(">", 0))

			Consistently(session.ExitCode, time.Duration(metricFrequency*2)*time.Second).Should(Equal(-1))
			contents, err = ioutil.ReadFile(logFilePath)
			Expect(err).NotTo(HaveOccurred())
			Expect(contents).NotTo(BeEmpty())
			newLineCount := bytes.Count(contents, []byte("\n"))
			Expect(newLineCount).Should(BeNumerically(">", firstLineCount))
		})

	})

	Describe("when the database is unreachable", func() {
		It("still logs the available=0 metric", func() {
			configFilepath = "./fixtures/bad-credentials-fixtures.yml"
			logFilePath := filepath.Join(tempDir, "metrics.log")
			runMainWithArgs("-l", logFilePath)

			Consistently(session.ExitCode, time.Duration(metricFrequency)*time.Second).Should(Equal(-1))
			contents, err := ioutil.ReadFile(logFilePath)
			Expect(err).NotTo(HaveOccurred())
			Expect(contents).NotTo(BeEmpty())
			contentsAsString := fmt.Sprintf("%s", contents)
			Expect(contentsAsString).To(ContainSubstring(`"key":"/p-mysql/available","value":0`))
			Expect(contentsAsString).NotTo(ContainSubstring("/p-mysql/innodb/buffer_pool_pages_free"))
		})
	})
})
