package main

import (
	"os"

	"io/ioutil"
	"strconv"
	"time"

	"code.cloudfoundry.org/clock"
	"code.cloudfoundry.org/lager"
	"fmt"
	"github.com/cloudfoundry-incubator/uaa-go-client"
	uaa_config "github.com/cloudfoundry-incubator/uaa-go-client/config"
	"github.com/pivotal-cf-experimental/notifications-client/notificationemailer"
	"github.com/pivotal-cf-experimental/replication-canary/alert"
	"github.com/pivotal-cf-experimental/replication-canary/canary"
	"github.com/pivotal-cf-experimental/replication-canary/clientcreator"
	"github.com/pivotal-cf-experimental/replication-canary/config"
	"github.com/pivotal-cf-experimental/replication-canary/database"
	"github.com/pivotal-cf-experimental/replication-canary/galera"
	"github.com/pivotal-cf-experimental/replication-canary/middleware"
	"github.com/pivotal-cf-experimental/replication-canary/switchboard"
	"net/http"
)

func main() {
	appConfig, err := config.NewConfig(os.Args)

	logger := appConfig.Logger

	if err != nil {
		logger.Fatal("Failed to read config", err, lager.Data{
			"config": sanitizeConfig(appConfig),
		})
	}

	err = appConfig.Validate()
	if err != nil {
		logger.Fatal("Failed to validate config", err, lager.Data{
			"config": sanitizeConfig(appConfig),
		})
	}

	pidfile, err := os.Create(appConfig.PidFile)
	if err != nil {
		logger.Fatal("Failed to create a file", err)
	}

	ioutil.WriteFile(pidfile.Name(), []byte(strconv.Itoa(os.Getpid())), 0644)

	logger.Info("Starting replication canary with configuration", lager.Data{
		"mysql": appConfig.MySQL,
	})

	skipSSLCertVerify := appConfig.SkipSSLValidation

	notificationsClient := &notificationemailer.Client{
		NotificationsDomain: "https://" + appConfig.Notifications.NotificationsDomain,
		SkipSSLCertVerify:   skipSSLCertVerify,
	}

	adminUaaConfig := &uaa_config.Config{
		ClientName:       appConfig.Notifications.AdminClientUsername,
		ClientSecret:     appConfig.Notifications.AdminClientSecret,
		UaaEndpoint:      "https://" + appConfig.Notifications.UAADomain,
		SkipVerification: skipSSLCertVerify,
	}

	adminUaaClient, err := uaa_go_client.NewClient(
		logger,
		adminUaaConfig,
		clock.NewClock(),
	)
	if err != nil {
		logger.Fatal("Failed to generate a UAA client", err)
	}

	err = clientcreator.CreateClient(adminUaaClient, appConfig)
	if err != nil {
		logger.Fatal("Failed to register UAA client", err)
	}

	uaaConfig := &uaa_config.Config{
		ClientName:       appConfig.Notifications.ClientUsername,
		ClientSecret:     appConfig.Notifications.ClientSecret,
		UaaEndpoint:      "https://" + appConfig.Notifications.UAADomain,
		SkipVerification: skipSSLCertVerify,
	}

	uaaClient, err := uaa_go_client.NewClient(
		logger,
		uaaConfig,
		clock.NewClock(),
	)
	if err != nil {
		logger.Fatal("Failed to generate a UAA client", err)
	}

	loggingAlerter := &alert.LoggingAlerter{
		Logger: logger,
	}

	emailingAlerter := &alert.EmailingAlerter{
		Logger:              logger,
		UAAClient:           uaaClient,
		NotificationsClient: notificationsClient,
		ToAddress:           appConfig.Notifications.ToAddress,
		SystemDomain:        appConfig.Notifications.SystemDomain,
		ClusterIdentifier:   appConfig.Notifications.ClusterIdentifier,
	}

	aggregateAlerter := alert.AggregateAlerter{
		loggingAlerter,
		emailingAlerter,
	}

	sc := []database.SwitchboardClient{}

	for _, url := range appConfig.Switchboard.URLs {
		switchboardClient := switchboard.NewClient(
			url,
			appConfig.Switchboard.Username,
			appConfig.Switchboard.Password,
			skipSSLCertVerify,
			logger,
		)

		sc = append(sc, switchboardClient)
		switchboardAlerter := &alert.SwitchboardAlerter{
			Logger:            logger,
			SwitchboardClient: switchboardClient,
			NoOp:              appConfig.NotifyOnly,
		}

		aggregateAlerter = append(aggregateAlerter, switchboardAlerter)

	}

	connFactory := database.NewConnectionFactoryFromConfig(appConfig, sc, logger)

	databaseSession := make(map[string]string)
	databaseSession["wsrep_sync_wait"] = "1"

	client := database.NewClient(databaseSession, logger)
	healthchecker := &galera.Client{
		Logger: logger,
	}
	bird := canary.NewCanary(client, healthchecker, time.Duration(appConfig.WriteReadDelay)*time.Second, logger)

	writeConn, err := connFactory.WriteConn()
	if err != nil {
		logger.Fatal("Canary setup failed", err)
	}

	err = bird.Setup(writeConn)
	if err != nil {
		logger.Fatal("Canary setup failed", err)
	}

	coalMiner := canary.NewCoalMiner(
		connFactory,
		bird,
		aggregateAlerter,
		logger,
	)

	go publishApi(coalMiner.StateMachine, appConfig.APIPort, appConfig.Canary.Username, appConfig.Canary.Password)

	ticker := time.NewTicker(time.Duration(appConfig.PollFrequency) * time.Second)
	logger.Info("ready to sing")
	coalMiner.LetSing(ticker.C)
}

func publishApi(stateMachine canary.StateMachine, port uint, username string, password string) {
	insecureStatusHandler := http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		healthy := stateMachine.GetState() == canary.NotUnhealthy
		fmt.Fprintf(w, "{ \"healthy\": %t }", healthy)
	})

	basicAuth := middleware.NewBasicAuth(username, password)
	secureStatusHandler := basicAuth.Wrap(insecureStatusHandler)

	http.Handle("/api/v1/status", secureStatusHandler)

	bindAddress := fmt.Sprintf(":%d", port)
	fmt.Println(fmt.Sprintf("Listening on: '%s'", bindAddress))

	http.ListenAndServe(bindAddress, nil)
}

func sanitizeConfig(cfg *config.Config) *config.Config {
	cfg.Canary.Password = "REDACTED"

	return cfg
}
