package main

import (
	"flag"
	"fmt"
	"github.com/pivotal-cf/mysql-diag/canaryclient"
	"github.com/pivotal-cf/mysql-diag/config"
	"github.com/pivotal-cf/mysql-diag/database"
	"github.com/pivotal-cf/mysql-diag/diagagentclient"
	"github.com/pivotal-cf/mysql-diag/msg"
	"github.com/pivotal-cf/mysql-diag/ui"
	"os"
)

const (
	defaultConfigPath = "/var/vcap/jobs/mysql-diag/config/mysql-diag-config.yml"
)

var configFilepath = flag.String("c", defaultConfigPath, "location of config file")

// returns true if the cluster needs bootstrap
func checkClusterStatus(mysqlConfig config.MysqlConfig) bool {
	clusterStateTable := ui.NewClusterStateTable(os.Stdout)

	rows := getClusterStatus(mysqlConfig)
	for _, row := range rows {
		n := row.node
		clusterStateTable.Add(n.Host, n.Name, n.UUID, row.status)
	}

	clusterStateTable.Render()

	statuses := make([]*database.GaleraStatus, len(rows))
	for i, row := range rows {
		statuses[i] = row.status
	}

	if database.DoWeNeedBoostrap(statuses) {
		return true
	} else {
		fmt.Println("I don't think bootstrap is necessary")
		return false
	}
}

type nodeStatus struct {
	node   config.MysqlNode
	status *database.GaleraStatus
}

func getClusterStatus(mysqlConfig config.MysqlConfig) []*nodeStatus {
	channel := make(chan nodeStatus, len(mysqlConfig.Nodes))

	for _, n := range mysqlConfig.Nodes {
		n := n

		go func() {
			intro := fmt.Sprintf("Checking cluster status of %s/%s at %s... ", n.Name, n.UUID, n.Host)
			fmt.Println(intro)

			ac := database.NewDatabaseClient(mysqlConfig.Connection(n))
			galeraStatus, err := ac.Status()
			if err != nil {
				msg.PrintfErrorIntro(intro, "%v", err)
			} else {
				fmt.Println(intro + "done")
			}

			channel <- nodeStatus{node: n, status: galeraStatus}
		}()
	}

	var nodeStatuses []*nodeStatus
	for i := 0; i < len(mysqlConfig.Nodes); i++ {
		ns := <-channel
		nodeStatuses = append(nodeStatuses, &ns)
	}

	return nodeStatuses
}

// Returns true if the canary is unhealthy. Otherwise it's either healthy or unknown.
func checkCanary(config *config.CanaryConfig) bool {
	if config == nil {
		fmt.Println("Canary not configured, skipping health check")
		return false
	}

	intro := "Checking canary status... "
	fmt.Println(intro)

	client := canaryclient.NewCanaryClient("127.0.0.1", config.ApiPort, *config)
	healthy, err := client.Status()
	if err != nil {
		msg.PrintfErrorIntro(intro, "%v", err)
		return false
	} else {
		if healthy {
			fmt.Println(intro + msg.Happy("healthy"))
			return false
		} else {
			fmt.Println(intro + msg.Alert("unhealthy"))
			return true
		}
	}
}

type nodeDiskInfo struct {
	node config.MysqlNode
	info *diagagentclient.InfoResponse
}

func getDisks(mysqlConfig config.MysqlConfig) []nodeDiskInfo {
	channel := make(chan nodeDiskInfo, len(mysqlConfig.Nodes))

	for _, n := range mysqlConfig.Nodes {
		n := n

		go func() {
			intro := fmt.Sprintf("Checking disk status of %s/%s at %s... ", n.Name, n.UUID, n.Host)
			fmt.Println(intro)

			client := diagagentclient.NewDiagAgentClient(n.Host, mysqlConfig.Agent.Port, mysqlConfig.Agent.Username, mysqlConfig.Agent.Password)
			info, err := client.Info()
			if err != nil {
				msg.PrintfErrorIntro(intro, "%v", err)
			} else {
				fmt.Println(intro + "done")
			}

			channel <- nodeDiskInfo{node: n, info: info}
		}()
	}

	var nodeDiskInfos []nodeDiskInfo
	for i := 0; i < len(mysqlConfig.Nodes); i++ {
		ns := <-channel
		nodeDiskInfos = append(nodeDiskInfos, ns)
	}

	return nodeDiskInfos
}

func checkDiskStatus(mysqlConfig config.MysqlConfig) {
	if mysqlConfig.Agent == nil {
		fmt.Println("Agent not configured, skipping disk check")
		return
	}

	rows := getDisks(mysqlConfig)

	if isAnyInfoPresent(rows) {
		diskInfoTable := ui.NewDiskInfoTable(os.Stdout)

		for _, row := range rows {
			n := row.node
			diskInfoTable.Add(n.Host, n.Name, n.UUID, row.info)
		}

		diskInfoTable.Render()
	} else {
		fmt.Println(msg.Alert("Unable to gather disk usage information, moving on. Run bosh vms --vitals for this information."))
	}
}

func isAnyInfoPresent(infos []nodeDiskInfo) bool {
	for _, info := range infos {
		if info.info != nil {
			return true
		}
	}
	return false
}

func downloadLogsExampleLine(hosts []string) string {
	line := "$ download-logs -d /tmp/output"
	for _, host := range hosts {
		line = line + " -n " + host
	}
	return line
}

func main() {
	flag.Parse()

	c, err := config.LoadFromFile(*configFilepath)
	if err != nil {
		msg.PrintfError("%v", err)
		os.Exit(1)
	}

	unhealthy := checkCanary(c.Canary)
	needsBootstrap := checkClusterStatus(c.Mysql)
	checkDiskStatus(c.Mysql)

	if unhealthy {
		fmt.Println(msg.Alert("\n[CRITICAL] The replication process is unhealthy. Writes are disabled."))
	}

	if needsBootstrap {
		fmt.Println(msg.Alert("\n[CRITICAL] You must boostrap the cluster. Follow these instructions: http://docs.pivotal.io/p-mysql/bootstrapping.html"))
	}

	if unhealthy || needsBootstrap {
		fmt.Println(msg.Alert("\n[CRITICAL] Run the download-logs command:\n") + downloadLogsExampleLine(c.HostsWithLogs()) + `
For full information about how to download and use the download-logs command see https://discuss.pivotal.io/hc/en-us/articles/221504408`)
		fmt.Print(msg.Alert("\n[WARNING] NOT RECOMMENDED") + `
Do not perform the following unless instructed by Pivotal Support:
- Do not scale down the cluster to one node then scale back. This puts user data at risk.
- Avoid "bosh recreate" and "bosh cck". These options remove logs on the VMs making it harder to diagnose cluster issues.

`)
	}
}
